from dataclasses import dataclass

import pytest

from dopynion.cards import Card, CardContainer, CardName, Village
from dopynion.data_model import (
    CardName as CardNameDataModel,
)
from dopynion.data_model import (
    CardNameAndHand,
    Cards,
    Hand,
    MoneyCardsInHand,
    PossibleCards,
)
from dopynion.game import Game
from dopynion.player import DefaultPlayerHooks, Player


def test_adventurer(empty_player: Player) -> None:
    player = empty_player

    player.hand.append_several(5, CardName.ADVENTURER)

    player.deck.append(CardName.ADVENTURER)
    player.deck.append(CardName.BUREAUCRAT)
    player.deck.append(CardName.GOLD)
    player.deck.append(CardName.COUNCILROOM)
    player.deck.append(CardName.CURSE)
    player.deck.append(CardName.SILVER)
    player.deck.append(CardName.DUCHY)
    player.deck.append(CardName.ESTATE)
    player.deck.append(CardName.COPPER)
    player.deck.append(CardName.SMITHY)

    player.start_turn()

    player.action(CardName.ADVENTURER)

    assert player.discard[0] == CardName.ADVENTURER
    assert player.discard[1] == CardName.BUREAUCRAT
    assert player.discard[2] == CardName.COUNCILROOM
    assert player.discard[3] == CardName.CURSE
    assert len(player.discard) == 4
    assert CardName.GOLD in player.hand
    assert CardName.SILVER in player.hand
    assert CardName.COPPER not in player.hand


def test_bureaucrat_enemy_with_victory_cards(
    game_with_two_players: tuple[Game, Player, Player],
) -> None:
    _, player, enemy = game_with_two_players

    player.hand.clear()
    player.hand.append_several(5, CardName.BUREAUCRAT)

    enemy.deck.clear()
    enemy.hand.clear()
    enemy.hand.append_several(5, CardName.DUCHY)

    player.start_turn()

    player.action(CardName.BUREAUCRAT)

    assert len(enemy.hand) == 4
    assert enemy.deck[0] == CardName.DUCHY


def test_bureaucrat_enemy_without_victory_cards(
    game_with_two_players: tuple[Game, Player, Player],
) -> None:
    _, player, enemy = game_with_two_players

    player.hand.clear()
    player.hand.append_several(5, CardName.BUREAUCRAT)

    enemy.deck.clear()
    enemy.hand.clear()
    enemy.hand.append_several(5, CardName.GOLD)

    player.start_turn()

    player.action(CardName.BUREAUCRAT)

    assert len(enemy.hand) == 5
    assert not enemy.deck


def test_bureaucrat_with_silver(empty_player: Player) -> None:
    player = empty_player
    player.hand.append_several(5, CardName.BUREAUCRAT)

    player.start_turn()

    player.action(CardName.BUREAUCRAT)

    assert len(player.hand) == 4
    assert player.deck[0] == CardName.SILVER


def test_bureaucrat_without_silver(
    game_with_two_players: tuple[Game, Player, Player],
) -> None:
    game, player, _ = game_with_two_players
    while CardName.SILVER in game.stock:
        game.stock.remove(CardName.SILVER)

    player.deck.clear()
    player.hand.clear()
    player.hand.append_several(5, CardName.BUREAUCRAT)

    player.start_turn()

    player.action(CardName.BUREAUCRAT)

    assert len(player.hand) == 4
    assert not player.deck


def test_cellar(empty_player: Player) -> None:
    class Hooks(DefaultPlayerHooks):
        def __init__(self, *args: tuple, **kwargs: dict) -> None:
            super().__init__(*args, **kwargs)
            self.nb_cards = 0

        def confirm_discard_card_from_hand(
            self,
            _decision_input: CardNameAndHand,
        ) -> bool:
            self.nb_cards += 1
            return self.nb_cards < 3

    player = empty_player
    player.hand.append_several(5, CardName.CELLAR)
    player.deck.append_several(2, CardName.GOLD)
    player.hooks = Hooks()

    player.start_turn()

    player.action(CardName.CELLAR)

    assert len(player.hand) == 4
    assert player.hand.gold_qty == 2
    assert len(player.discard) == 2


def test_chancellor_dont_discard_deck(empty_player: Player) -> None:
    class Hooks(DefaultPlayerHooks):
        def confirm_discard_deck(self) -> bool:  # noqa: PLR6301
            return False

    player = empty_player
    player.hand.append_several(5, CardName.CHANCELLOR)
    player.deck.append_several(3, CardName.GOLD)
    player.hooks = Hooks()

    player.start_turn()

    player.action(CardName.CHANCELLOR)

    assert len(player.hand) == 4
    assert len(player.deck) == 3
    assert len(player.discard) == 0


def test_chancellor_discard_deck(empty_player: Player) -> None:
    class Hooks(DefaultPlayerHooks):
        def confirm_discard_deck(self) -> bool:  # noqa: PLR6301
            return True

    player = empty_player
    player.hand.append_several(5, CardName.CHANCELLOR)
    player.deck.append_several(3, CardName.GOLD)
    player.hooks = Hooks()

    player.start_turn()

    player.action(CardName.CHANCELLOR)

    assert len(player.hand) == 4
    assert len(player.deck) == 0
    assert len(player.discard) == 3


def test_chapel_trash_two_cards(empty_player: Player) -> None:
    class Hooks(DefaultPlayerHooks):
        def __init__(self, *args: tuple, **kwargs: dict) -> None:
            super().__init__(*args, **kwargs)
            self.nb_cards = 0

        def confirm_trash_card_from_hand(
            self,
            _decision_input: CardNameAndHand,
        ) -> bool:
            self.nb_cards += 1
            return self.nb_cards < 3

    player = empty_player
    player.hand.append_several(5, CardName.CHAPEL)
    player.hooks = Hooks()

    player.start_turn()

    player.action(CardName.CHAPEL)

    assert len(player.hand) == 2
    assert len(player.played_cards) == 1
    assert len(player.deck) == 0
    assert len(player.discard) == 0


def test_chapel_trash_as_many_cards_as_possible_that_is_four(
    empty_player: Player,
) -> None:
    class Hooks(DefaultPlayerHooks):
        def confirm_trash_card_from_hand(  # noqa: PLR6301
            self,
            _decision_input: CardNameAndHand,
        ) -> bool:
            return True

    player = empty_player
    player.hand.append_several(15, CardName.CHAPEL)
    player.hooks = Hooks()

    player.start_turn()

    player.action(CardName.CHAPEL)

    assert len(player.hand) == 10
    assert len(player.played_cards) == 1
    assert len(player.deck) == 0
    assert len(player.discard) == 0


def test_feast_trash(empty_player: Player) -> None:
    player = empty_player
    player.hand.append_several(5, CardName.FEAST)
    player.start_turn()

    player.action(CardName.FEAST)

    assert len(player.played_cards) == 0


def test_feast_trash_good_cardname(empty_player: Player) -> None:
    player = empty_player
    player.hand.append_several(4, CardName.FEAST)
    player.hand.append(CardName.FESTIVAL)
    player.start_turn()
    player.action(CardName.FESTIVAL)

    player.action(CardName.FEAST)

    assert len(player.played_cards) == 1
    assert CardName.FESTIVAL in player.played_cards


def test_feast_choose_a_card(empty_player: Player) -> None:
    class Hooks(DefaultPlayerHooks):
        def choose_card_to_receive_in_discard(  # noqa: PLR6301
            self,
            decision_input: PossibleCards,
        ) -> CardNameDataModel:
            return decision_input.possible_cards[0]

    player = empty_player
    player.hooks = Hooks()
    player.hand.append_several(5, CardName.FEAST)
    player.start_turn()

    player.action(CardName.FEAST)

    assert len(player.discard) == 1


def test_council_room(game_with_two_players: tuple[Game, Player, Player]) -> None:
    _, player, enemy = game_with_two_players
    player.hand.clear()
    player.hand.append_several(5, CardName.COUNCILROOM)
    enemy.hand.clear()

    player.start_turn()

    player.action(CardName.COUNCILROOM)

    assert len(player.hand) == 4 + Card.types[CardName.COUNCILROOM].more_cards_from_deck
    assert len(enemy.hand) == 1


def test_gardens_score_39_cards(empty_player: Player) -> None:
    player = empty_player
    player.deck.append_several(2, CardName.GARDENS)
    player.deck.append_several(37, CardName.COPPER)

    score = player.score()
    assert score["gardens_qty"] == 2
    assert score["score"] == 6


def test_gardens_score_40_cards(empty_player: Player) -> None:
    player = empty_player
    player.deck.append_several(2, CardName.GARDENS)
    player.deck.append_several(38, CardName.COPPER)

    score = player.score()
    assert score["gardens_qty"] == 2
    assert score["score"] == 8


def test_library(empty_player: Player) -> None:
    class Hooks(DefaultPlayerHooks):
        nb_max_skip = 3

        def __init__(self, *args: tuple, **kwargs: dict) -> None:
            super().__init__(*args, **kwargs)
            self.nb_cards = 0

        def skip_card_reception_in_hand(
            self,
            _decision_input: CardNameAndHand,
        ) -> bool:
            self.nb_cards += 1
            return self.nb_cards <= self.nb_max_skip

    player = empty_player
    player.hooks = Hooks()
    player.deck.append_several(7, CardName.VILLAGE)
    player.deck.append_several(50, CardName.COPPER)
    player.hand.append(CardName.GOLD)
    player.hand.append(CardName.LIBRARY)
    player.start_turn()

    player.action(CardName.LIBRARY)

    assert len(player.hand) == 7
    assert len(player.discard) == player.hooks.nb_max_skip


def test_militia(game_with_two_players: tuple[Game, Player, Player]) -> None:
    _, player, enemy = game_with_two_players
    player.hand.clear()
    player.hand.append_several(5, CardName.MILITIA)
    enemy.hand.clear()
    enemy.hand.append_several(5, CardName.MILITIA)

    player.start_turn()

    player.action(CardName.MILITIA)

    assert len(player.hand) == 4
    assert len(enemy.hand) == 3
    assert len(enemy.discard) == 2


def test_mine_don_t_trash(empty_player: Player) -> None:
    player = empty_player
    player.hand.append(CardName.COPPER)
    player.hand.append(CardName.MINE)
    player.start_turn()

    player.action(CardName.MINE)

    assert len(player.hand) == 1
    assert player.hand[0] == CardName.COPPER


def test_mine_trash_copper(empty_player: Player) -> None:
    class Hooks(DefaultPlayerHooks):
        def trash_money_card_for_better_money_card(  # noqa: PLR6301
            self,
            _decision_input: MoneyCardsInHand,
        ) -> CardNameDataModel | None:
            return CardName.COPPER

    player = empty_player
    player.hooks = Hooks()
    player.hand.append(CardName.COPPER)
    player.hand.append(CardName.MINE)
    player.start_turn()

    player.action(CardName.MINE)

    assert len(player.hand) == 1
    assert player.hand[0] == CardName.SILVER


def test_mine_trash_silver(empty_player: Player) -> None:
    class Hooks(DefaultPlayerHooks):
        def trash_money_card_for_better_money_card(  # noqa: PLR6301
            self,
            _decision_input: MoneyCardsInHand,
        ) -> CardNameDataModel | None:
            return CardName.SILVER

    player = empty_player
    player.hooks = Hooks()
    player.hand.append(CardName.SILVER)
    player.hand.append(CardName.MINE)
    player.start_turn()

    player.action(CardName.MINE)

    assert len(player.hand) == 1
    assert player.hand[0] == CardName.GOLD


def test_mine_trash_gold(empty_player: Player) -> None:
    class Hooks(DefaultPlayerHooks):
        def trash_money_card_for_better_money_card(  # noqa: PLR6301
            self,
            _decision_input: MoneyCardsInHand,
        ) -> CardNameDataModel | None:
            return CardName.GOLD

    player = empty_player
    player.hooks = Hooks()
    player.hand.append(CardName.GOLD)
    player.hand.append(CardName.MINE)
    player.start_turn()

    player.action(CardName.MINE)

    assert len(player.hand) == 1
    assert player.hand[0] == CardName.PLATINUM


def test_mine_trash_platinum(empty_player: Player) -> None:
    class Hooks(DefaultPlayerHooks):
        def trash_money_card_for_better_money_card(  # noqa: PLR6301
            self,
            _decision_input: MoneyCardsInHand,
        ) -> CardNameDataModel | None:
            return CardName.PLATINUM

    player = empty_player
    player.hooks = Hooks()
    player.hand.append(CardName.PLATINUM)
    player.hand.append(CardName.MINE)
    player.start_turn()

    player.action(CardName.MINE)

    assert len(player.hand) == 1
    assert player.hand[0] == CardName.PLATINUM


# TODO trash don't exist
# TODO check dans les hooks quand ils piochent dans une liste que l'élément choisi est
# bien dans la liste


def test_money_lender_refuse(empty_player: Player) -> None:
    player = empty_player
    player.hand.append_several(2, CardName.COPPER)
    player.hand.append_several(3, CardName.MONEYLENDER)
    player.start_turn()

    player.action(CardName.MONEYLENDER)

    assert len(player.hand) == 4
    assert player.hand.copper_qty == 2


def test_money_lender_accept(empty_player: Player) -> None:
    class Hooks(DefaultPlayerHooks):
        def confirm_trash_card_from_hand(  # noqa: PLR6301
            self,
            _decision_input: CardNameAndHand,
        ) -> bool:
            return True

    player = empty_player
    player.hooks = Hooks()
    player.hand.append_several(2, CardName.COPPER)
    player.hand.append_several(3, CardName.MONEYLENDER)
    player.start_turn()

    player.action(CardName.MONEYLENDER)

    assert len(player.deck) == 0
    assert len(player.discard) == 0
    assert len(player.played_cards) == 1
    assert player.hand.copper_qty == 1
    assert len(player.hand) == 4 - 1
    assert player.money == 3  # hand.money is not in money


def test_remodel(empty_player: Player) -> None:
    class Hooks(DefaultPlayerHooks):
        def trash_card_from_hand(  # noqa: PLR6301
            self,
            _hand: Hand,
        ) -> CardNameDataModel:
            return CardName.COPPER

        def choose_card_to_receive_in_discard(  # noqa: PLR6301
            self,
            decision_input: PossibleCards,
        ) -> CardNameDataModel:
            for card_name in decision_input.possible_cards:
                assert (
                    Card.types[CardName[card_name.upper()]].cost
                    <= Card.types[CardName.COPPER].cost + 2
                )
            return decision_input.possible_cards[0]

    player = empty_player
    player.hooks = Hooks()
    player.hand.append_several(1, CardName.COPPER)
    player.hand.append_several(3, CardName.SILVER)
    player.hand.append_several(1, CardName.REMODEL)
    player.start_turn()

    player.action(CardName.REMODEL)

    assert len(player.deck) == 0
    assert len(player.played_cards) == 1
    assert len(player.hand) == 3
    assert len(player.discard) == 1  # copper is trashed, but received one card


def test_witch(game_with_two_players: tuple[Game, Player, Player]) -> None:
    _, player, enemy = game_with_two_players
    player.hand.clear()
    player.hand.append_several(5, CardName.WITCH)
    enemy.hand.clear()
    enemy.hand.append_several(5, CardName.WITCH)

    player.start_turn()

    player.action(CardName.WITCH)

    assert len(player.hand) == 4 + Card.types[CardName.WITCH].more_cards_from_deck
    assert len(enemy.hand) == 5
    assert CardName.CURSE in enemy.discard


def test_workshop(empty_player: Player) -> None:
    class Hooks(DefaultPlayerHooks):
        def choose_card_to_receive_in_discard(  # noqa: PLR6301
            self,
            decision_input: PossibleCards,
        ) -> CardNameDataModel:
            for card_name in decision_input.possible_cards:
                assert Card.types[CardName[card_name.upper()]].cost <= 4
            return decision_input.possible_cards[0]

    player = empty_player
    player.hooks = Hooks()
    player.hand.append_several(5, CardName.WORKSHOP)

    player.start_turn()

    player.action(CardName.WORKSHOP)

    assert len(player.discard) == 1
    assert len(player.hand) == 4


def test_swap(empty_player: Player) -> None:
    class Hooks(DefaultPlayerHooks):
        def confirm_trash_card_from_hand(  # noqa: PLR6301
            self,
            _decision_input: CardNameAndHand,
        ) -> bool:
            return True

    player = empty_player
    player.hooks = Hooks()
    player.hand.append_several(5, CardName.SWAP)

    player.start_turn()
    swap_stock_before = player.game.stock.swap_qty

    player.action(CardName.SWAP)

    swap_stock_after = player.game.stock.swap_qty
    assert swap_stock_after == swap_stock_before + 1
    assert len(player.hand) == 3  # 4 - 1
    assert len(player.discard) == 1


def test_swap_no_action(empty_player: Player) -> None:
    class Hooks(DefaultPlayerHooks):
        def confirm_trash_card_from_hand(  # noqa: PLR6301
            self,
            _decision_input: CardNameAndHand,
        ) -> bool:
            return True

    player = empty_player
    player.hooks = Hooks()
    player.hand.append(CardName.SWAP)
    player.hand.append_several(4, CardName.GOLD)

    player.start_turn()

    player.action(CardName.SWAP)

    assert len(player.hand) == 4
    assert len(player.discard) == 0


def test_artificer(empty_player: Player) -> None:
    class Hooks(DefaultPlayerHooks):
        def __init__(self, *args: tuple, **kwargs: dict) -> None:
            super().__init__(*args, **kwargs)
            self.nb_cards = 0

        def confirm_discard_card_from_hand(
            self,
            _decision_input: CardNameAndHand,
        ) -> bool:
            self.nb_cards += 1
            return self.nb_cards <= 3

        def choose_card_to_receive_in_deck(  # noqa: PLR6301
            self,
            decision_input: PossibleCards,
        ) -> CardNameDataModel:
            return CardName.VILLAGE

    player = empty_player
    player.game.stock.append(CardName.VILLAGE)
    player.hooks = Hooks()
    player.hand.append(CardName.ARTIFICER)
    player.hand.append_several(4, CardName.GOLD)

    player.start_turn()

    player.action(CardName.ARTIFICER)

    assert len(player.hand) == 4 - 3
    assert len(player.discard) == 3
    assert len(player.deck) == 1


def test_distant_shore(empty_player: Player) -> None:
    player = empty_player
    player.hand.append_several(5, CardName.DISTANTSHORE)

    player.start_turn()

    player.action(CardName.DISTANTSHORE)

    assert len(player.hand) == 4
    assert CardName.ESTATE in player.discard


def test_distant_shore_no_estate(empty_player: Player) -> None:
    player = empty_player
    player.hand.append_several(5, CardName.DISTANTSHORE)
    while CardName.ESTATE in player.game.stock:
        player.game.stock.remove(CardName.ESTATE)

    player.start_turn()

    player.action(CardName.DISTANTSHORE)

    assert len(player.hand) == 4
    assert len(player.discard) == 0


def test_hireling_more_cards_at_start(empty_player: Player) -> None:
    player = empty_player
    player.deck.append_several(10, CardName.GOLD)
    player.hand.append_several(4, CardName.COPPER)
    player.hand.append(CardName.HIRELING)

    player.start_turn()

    player.action(CardName.HIRELING)
    assert CardName.HIRELING not in player.played_cards
    assert CardName.HIRELING not in player.discard

    player.end_turn()
    player.start_turn()
    assert len(player.hand) == 6


@dataclass
class CardParameter:
    card_name: CardName
    more_purchase: int
    more_actions: int
    more_money: int
    more_cards: int


@pytest.mark.parametrize(
    ("card_param"),
    [
        CardParameter(
            CardName.ARTIFICER,
            more_purchase=0,
            more_actions=1,
            more_money=1,
            more_cards=1,
        ),
        CardParameter(
            CardName.CELLAR,
            more_purchase=0,
            more_actions=1,
            more_money=0,
            more_cards=0,
        ),
        CardParameter(
            CardName.CHANCELLOR,
            more_purchase=0,
            more_actions=0,
            more_money=2,
            more_cards=0,
        ),
        CardParameter(
            CardName.COUNCILROOM,
            more_purchase=1,
            more_actions=0,
            more_money=0,
            more_cards=4,
        ),
        CardParameter(
            CardName.DISTANTSHORE,
            more_purchase=0,
            more_actions=1,
            more_money=0,
            more_cards=2,
        ),
        CardParameter(
            CardName.FESTIVAL,
            more_purchase=1,
            more_actions=2,
            more_money=2,
            more_cards=0,
        ),
        CardParameter(
            CardName.HIRELING,
            more_purchase=0,
            more_actions=0,
            more_money=0,
            more_cards=1,
        ),
        CardParameter(
            CardName.LABORATORY,
            more_purchase=0,
            more_actions=1,
            more_money=0,
            more_cards=2,
        ),
        CardParameter(
            CardName.MARKET,
            more_purchase=1,
            more_actions=1,
            more_money=1,
            more_cards=1,
        ),
        CardParameter(
            CardName.MILITIA,
            more_purchase=0,
            more_actions=0,
            more_money=2,
            more_cards=0,
        ),
        CardParameter(
            CardName.SMITHY,
            more_purchase=0,
            more_actions=0,
            more_money=0,
            more_cards=3,
        ),
        CardParameter(
            CardName.SWAP,
            more_purchase=0,
            more_actions=1,
            more_money=0,
            more_cards=1,
        ),
        CardParameter(
            CardName.VILLAGE,
            more_purchase=0,
            more_actions=2,
            more_money=0,
            more_cards=1,
        ),
        CardParameter(
            CardName.WITCH,
            more_purchase=0,
            more_actions=0,
            more_money=0,
            more_cards=2,
        ),
        CardParameter(
            CardName.WOODCUTTER,
            more_purchase=1,
            more_actions=0,
            more_money=2,
            more_cards=0,
        ),
    ],
)
def test_basic_cards(
    player_with_action_card: Player,
    card_param: CardParameter,
) -> None:
    player = player_with_action_card
    player.hand.append(card_param.card_name)
    player.start_turn()
    old_purchases_left = player.purchases_left
    old_actions_left = player.actions_left
    old_money = player.money
    old_nb_cards = len(player.hand)

    player.action(card_param.card_name)

    assert Card.types[card_param.card_name].is_action
    assert player.purchases_left == old_purchases_left + card_param.more_purchase
    assert player.actions_left == (old_actions_left - 1) + card_param.more_actions
    assert player.money == old_money + card_param.more_money
    assert len(player.hand) == (old_nb_cards - 1) + card_param.more_cards


def test_card_coverage(player: Player) -> None:
    card = Village()
    print(card, card == Village, hash(card), player.hand)


def test_cards_quantities() -> None:
    cards = CardContainer()
    cards.append(CardName.VILLAGE)
    cards.append(CardName.DUCHY)
    cards.remove(CardName.VILLAGE)
    assert cards.state == Cards(quantities={CardName.DUCHY: 1})



import datetime
from pathlib import Path
from typing import Literal

from dopynion.data_model import (
    ActionRecord,
    Cards,
    ErrorRecord,
    Game,
    GameRecord,
    HookCallArgs,
    HookCallRecord,
    HookCallResult,
    HookResultRecord,
    PlayerTurnRecord,
)
from dopynion.player import Player

records_dir = Path.cwd() / "games"
records_dir.mkdir(parents=True, exist_ok=True)


class Record:
    def __init__(self) -> None:
        now = datetime.datetime.now(tz=datetime.UTC)
        now_str = now.strftime("%Y_%m_%d__%H_%M_%S_%f")
        self._file = records_dir / f"game__{now_str}.dop"
        if self._file.exists():
            msg = f"game record is already created ({self._file})"
            raise ValueError(msg)
        self._file.write_text("", encoding="utf-8")
        self._game_record = GameRecord(date=now, stock=Cards())
        self.save(Game(finished=False, players=[], stock=Cards()))

    @staticmethod
    def load(path: Path) -> GameRecord:
        return GameRecord.model_validate_json(path.read_text(encoding="utf-8"))

    def save(self, game: Game) -> Path:
        for player in game.players:
            self._game_record.scores[player.name] = player.score
        self._file.write_text(self._game_record.model_dump_json(indent=None))
        return self._file

    def start_turn(self) -> None:
        self._game_record.turns.append(PlayerTurnRecord())

    def add_stock(self, stock: Cards) -> None:
        self._game_record.stock = stock

    def add_action(self, action: str, player: Player) -> None:
        turn = self._game_record.turns[-1]
        action_record = ActionRecord(
            action=action,
            player=player.state,
            score=player.score()["score"],
        )
        turn.actions.append(action_record)

    def add_error(self, error: str, player: Player) -> None:
        self._add_error(error, player, "error")

    def add_warning(self, error: str, player: Player) -> None:
        self._add_error(error, player, "warning")

    def _add_error(
        self,
        error: str,
        player: Player,
        type_: Literal["error", "warning"],
    ) -> None:
        if not self._game_record.turns:
            self._game_record.turns.append(PlayerTurnRecord())
        turn = self._game_record.turns[-1]
        error_record = ErrorRecord(error=error, player=player.state, type=type_)
        turn.actions.append(error_record)

    def add_hook_call(self, player: Player, name: str, args: HookCallArgs) -> None:
        turn = self._game_record.turns[-1]
        turn.actions.append(HookCallRecord(name=name, player=player.state, args=args))

    def add_hook_result(self, player: Player, result: HookCallResult) -> None:
        turn = self._game_record.turns[-1]
        turn.actions.append(HookResultRecord(player=player.state, result=result))



from datetime import datetime
from enum import StrEnum
from typing import Literal

from pydantic import BaseModel, Field


class CardName(StrEnum):  # Create with a metaclass
    ADVENTURER = "adventurer"
    ARTIFICER = "artificer"
    BUREAUCRAT = "bureaucrat"
    CELLAR = "cellar"
    CHANCELLOR = "chancellor"
    CHAPEL = "chapel"
    COLONY = "colony"
    COPPER = "copper"
    COUNCILROOM = "councilroom"
    CURSE = "curse"
    DISTANTSHORE = "distantshore"
    DUCHY = "duchy"
    ESTATE = "estate"
    FEAST = "feast"
    FESTIVAL = "festival"
    GARDENS = "gardens"
    GOLD = "gold"
    HIRELING = "hireling"
    LABORATORY = "laboratory"
    LIBRARY = "library"
    MAGNATE = "magnate"
    MARKET = "market"
    MILITIA = "militia"
    MINE = "mine"
    MONEYLENDER = "moneylender"
    NONE = "NONE"
    PLATINUM = "platinum"
    PROVINCE = "province"
    REMODEL = "remodel"
    SILVER = "silver"
    SMITHY = "smithy"
    SWAP = "swap"
    VILLAGE = "village"
    WITCH = "witch"
    WOODCUTTER = "woodcutter"
    WORKSHOP = "workshop"

    def __repr__(self) -> str:
        return self.value.title()


class Cards(BaseModel):
    quantities: dict[CardName, int] = Field(default_factory=dict)


class Player(BaseModel):
    name: str
    hand: Cards | None
    score: int


class Game(BaseModel):
    finished: bool
    players: list[Player]
    stock: Cards


class ActionRecord(BaseModel):
    action: str
    player: Player
    score: int


class ErrorRecord(BaseModel):
    error: str
    player: Player
    type: Literal["error", "warning"] = "error"


class CardNameAndHand(BaseModel):
    card_name: CardName
    hand: list[CardName]


class Hand(BaseModel):
    hand: list[CardName]


class PossibleCards(BaseModel):
    possible_cards: list[CardName]


class MoneyCardsInHand(BaseModel):
    money_in_hand: list[CardName]


HookCallArgs = CardNameAndHand | Hand | PossibleCards | MoneyCardsInHand | None
HookCallResult = CardName | bool | None


class HookCallRecord(BaseModel):
    name: str
    player: Player
    args: HookCallArgs


class HookResultRecord(BaseModel):
    player: Player
    result: HookCallResult


class PlayerTurnRecord(BaseModel):
    actions: list[ActionRecord | ErrorRecord | HookCallRecord | HookResultRecord] = (
        Field(default_factory=list)
    )


class GameRecord(BaseModel):
    date: datetime
    stock: Cards
    turns: list[PlayerTurnRecord] = Field(default_factory=list)
    scores: dict[str, int] = Field(default_factory=dict)



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from collections.abc import Callable
from enum import Enum
from typing import TYPE_CHECKING, TypeVar, cast, overload

from dopynion.cards import Card, CardContainer
from dopynion.data_model import (
    CardName,
    CardNameAndHand,
    Hand,
    HookCallArgs,
    HookCallResult,
    MoneyCardsInHand,
    PossibleCards,
)
from dopynion.data_model import Player as PlayerData
from dopynion.exceptions import (
    ActionDuringBuyError,
    InvalidActionError,
    InvalidBuyError,
    InvalidDiscardError,
    NotEnoughMoneyError,
)

if TYPE_CHECKING:
    from collections.abc import Generator

    import dopynion.game

logger = logging.getLogger(__name__)


class State(Enum):
    ACTION = 1
    BUY = 2
    ADJUST = 3


HookArg = TypeVar("HookArg", bound=HookCallArgs)
HookRet = TypeVar("HookRet", bound=HookCallResult)


class PlayerHooks(ABC):
    @abstractmethod
    def confirm_discard_card_from_hand(
        self,
        decision_input: CardNameAndHand,
    ) -> bool:
        pass

    @abstractmethod
    def confirm_trash_card_from_hand(
        self,
        decision_input: CardNameAndHand,
    ) -> bool:
        pass

    @abstractmethod
    def discard_card_from_hand(
        self,
        decision_input: Hand,
    ) -> CardName:
        pass

    @abstractmethod
    def trash_card_from_hand(
        self,
        decision_input: Hand,
    ) -> CardName:
        pass

    @abstractmethod
    def confirm_discard_deck(self) -> bool:
        pass

    @abstractmethod
    def choose_card_to_receive_in_discard(
        self,
        decision_input: PossibleCards,
    ) -> CardName:
        pass

    @abstractmethod
    def choose_card_to_receive_in_deck(
        self,
        decision_input: PossibleCards,
    ) -> CardName:
        pass

    @abstractmethod
    def skip_card_reception_in_hand(
        self,
        decision_input: CardNameAndHand,
    ) -> bool:
        pass

    @abstractmethod
    def trash_money_card_for_better_money_card(
        self,
        decision_input: MoneyCardsInHand,
    ) -> CardName | None:
        pass


class DefaultPlayerHooks(PlayerHooks):
    def confirm_discard_card_from_hand(  # noqa: PLR6301
        self,
        _decision_input: CardNameAndHand,
    ) -> bool:
        return False

    def confirm_trash_card_from_hand(  # noqa: PLR6301
        self,
        _decision_input: CardNameAndHand,
    ) -> bool:
        return False

    def discard_card_from_hand(  # noqa: PLR6301
        self,
        decision_input: Hand,
    ) -> CardName:
        return decision_input.hand[0]

    def trash_card_from_hand(  # noqa: PLR6301
        self,
        decision_input: Hand,
    ) -> CardName:
        return decision_input.hand[0]

    def confirm_discard_deck(self) -> bool:  # noqa: PLR6301
        return False

    def choose_card_to_receive_in_discard(  # noqa: PLR6301
        self,
        decision_input: PossibleCards,
    ) -> CardName:
        return decision_input.possible_cards[0]

    def choose_card_to_receive_in_deck(  # noqa: PLR6301
        self,
        decision_input: PossibleCards,
    ) -> CardName:
        return decision_input.possible_cards[0]

    def skip_card_reception_in_hand(  # noqa: PLR6301
        self,
        _decision_input: CardNameAndHand,
    ) -> bool:
        return False

    def trash_money_card_for_better_money_card(  # noqa: PLR6301
        self,
        _decision_input: MoneyCardsInHand,
    ) -> CardName | None:
        return None


class Player:
    def __init__(self, name: str) -> None:
        self.game: dopynion.game.Game = None  # type: ignore[assignment]
        self.name = name
        self.deck = CardContainer()
        self.deck.append_several(7, CardName.COPPER)
        self.deck.append_several(3, CardName.ESTATE)
        self.deck.shuffle()
        self.hand = CardContainer()
        self.discard = CardContainer()
        self.played_cards = CardContainer()
        self.actions_left: int = 0
        self.purchases_left: int = 0
        self.money: int = 0
        self.playing = False
        self.state_machine: State = State.ACTION
        self.eliminated = False
        self.nb_cards_in_hand_at_turn_start = 5
        self.hooks: PlayerHooks = DefaultPlayerHooks()
        self._adjust()

    def __repr__(self) -> str:
        ret = f"Player: {self.name}\n"
        ret += f" - actions left: {self.actions_left}, "
        ret += f"purchases left: {self.purchases_left}, "
        ret += f"money left: {self.money + self.hand.money}\n"
        ret += f" - hand: {self.hand}\n"
        ret += f" - played cards: {self.played_cards}\n"
        ret += f" - deck: {self.deck}\n"
        ret += f" - discard: {self.discard}\n"
        return ret

    def _check_for_action_to_buy_transition(self) -> None:
        if not self.hand.contains_action() or not self.actions_left:
            self.actions_left = 0
            self.state_machine = State.BUY
            logger.debug("go from action to buy")

    def _check_for_buy_to_adjust_transition(self) -> None:
        if not self.purchases_left:
            self.purchases_left = 0
            self.state_machine = State.ADJUST
            logger.debug("go from buy to adjust")

    def start_turn(self) -> None:
        logger.debug("start turn (%s, %d)", self.name, id(self))
        self.game.record.start_turn()
        self.playing = True
        self.state_machine = State.ACTION
        self.actions_left = 1
        self.purchases_left = 1
        self.money = 0
        self._check_for_action_to_buy_transition()

    def end_turn(self) -> None:
        self.game.record.add_action("END OF TURN", self)
        self._adjust()
        self.playing = False
        logger.debug("end turn")

    def take_one_card_from_deck(self) -> CardName | None:
        if not self.deck:
            self.discard.empty_to(self.deck)
            self.deck.shuffle()
        if not self.deck:
            return None
        return self.deck.pop(0)

    def _adjust(self) -> None:
        self.played_cards.empty_to(self.discard)
        self.hand.empty_to(self.discard)
        for _ in range(self.nb_cards_in_hand_at_turn_start):
            card_name = self.take_one_card_from_deck()
            if card_name is not None:
                self.hand.append(card_name)
        self.state_machine = State.ADJUST

    def _prepare_money(self, money: int) -> None:
        money_cards = self.hand.money_cards
        money_cards.sort(key=lambda card_name: Card.types[card_name].money)
        while self.money < money:
            if not money_cards:
                break
            money_card = money_cards.pop(0)
            self.money += Card.types[money_card].money
            self.hand.remove(money_card)
            self.played_cards.append(money_card)

    def buy(self, card_name: CardName) -> None:
        logger.debug("> BUY %s", card_name)
        self.game.record.add_action(f"BUY {card_name}", self)
        if not self.purchases_left:
            msg = "No more buy available"
            self.game.record.add_error(msg, self)
            raise InvalidBuyError(msg)
        quantity = getattr(self.game.stock, card_name + "_qty")
        if not quantity:
            self.game.record.add_error(f"Invalid buy, no {card_name} in stock", self)
            raise InvalidBuyError(card_name)
        card = Card.types[card_name]
        if self.money + self.hand.money < card.cost:
            self.game.record.add_error("Invalid buy, not enough money", self)
            raise NotEnoughMoneyError
        self._prepare_money(card.cost)
        self.money -= card.cost
        self.game.stock.remove(card_name)
        self.discard.append(card_name)
        self.purchases_left -= 1
        self._check_for_buy_to_adjust_transition()

    def action(self, card_name: CardName) -> None:
        logger.debug("> ACTION %s", card_name)
        self.game.record.add_action(f"ACTION {card_name}", self)
        if self.state_machine != State.ACTION:
            logger.debug("actions_left %d", self.actions_left)
            logger.debug(self.state_machine.name)
            logger.debug(self.state)
            self.game.record.add_error("Tried action during buy phase", self)
            raise ActionDuringBuyError(card_name)
        if card_name not in self.hand:
            logger.debug(self.state)
            self.game.record.add_error(f"Invalid action, {card_name} not in hand", self)
            raise InvalidActionError(card_name)
        self.actions_left -= 1
        self.hand.remove(card_name)
        self.played_cards.append(card_name)
        Card.types[card_name].action(self)
        self._check_for_action_to_buy_transition()

    @property
    def state(self) -> PlayerData:
        score = self.score()["score"]
        state = PlayerData(name=self.name, hand=None, score=score)
        if self.playing and not self.eliminated:
            state.hand = self.hand.state
        return state

    def eliminate(self) -> None:
        self.eliminated = True
        self.game.record.add_error("Elimination", self)

    def score(self) -> dict:
        cards = self.hand + self.discard + self.deck
        if self.eliminated:
            return {"score": -10000}
        return {
            "colony_qty": cards.colony_qty,
            "province_qty": cards.province_qty,
            "duchy_qty": cards.duchy_qty,
            "estate_qty": cards.estate_qty,
            "curse_qty": cards.curse_qty,
            "gardens_qty": cards.gardens_qty,
            "score": sum(Card.types[card_name].victory_points for card_name in cards)
            + cards.gardens_qty * (len(cards) // 10),
        }

    def discard_one_card_from_hand(self, card_name: CardName) -> None:
        if card_name not in self.hand:
            msg = f"{card_name} not in player hand"
            self.game.record.add_error("Invalid discard, " + msg, self)
            raise InvalidDiscardError(msg)
        self.hand.remove(card_name)
        self.discard.append(card_name)

    def other_players(self) -> Generator[Player, None, None]:
        for other_player in self.game.players:
            if other_player != self and not other_player.eliminated:
                yield other_player

    @overload
    def use_hook(self, hook: Callable[[], HookRet]) -> HookRet: ...
    @overload
    def use_hook(
        self,
        hook: Callable[[HookArg], HookRet],
        args: HookArg,
    ) -> HookRet: ...

    def use_hook(
        self,
        hook: Callable[[], HookRet] | Callable[[HookArg], HookRet],
        args: HookArg | None = None,
    ) -> HookRet:
        self.game.record.add_hook_call(self, hook.__name__, args)
        if args is None:
            zero_arg_hook = cast(Callable[[], HookRet], hook)
            result = zero_arg_hook()
        else:
            arg_hook = cast(Callable[[HookArg], HookRet], hook)
            result = arg_hook(args)
        self.game.record.add_hook_result(self, result)
        return result
